# Copyright 2011 by Andrew Scullion

# Permission to use, copy, modify, and distribute this software for
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

# Name :          Voxelize
# Description :   Transform a solid into little cubes
# Author :        TheDro
# Usage :         Select a group which contains a solid (no faces inside
# 								the solid for proper behavior), run the script, choose
#								  the size of each cube and whether a full solid of cubes
# 								or a shell of cubes should be made. Click OK to confirm.
# Warning :       The process is still very slow and may cause Sketchup to
# 								"not respond" even though the script is running in the 
# 								background. Apply this to relatively simple geometry
#									until future releases.
# Date :          2011/04/04
# History:
#		beta 01 : First released version ( rev a: module wrapped )
#
#   beta 02 : The algorithm to determine the position of the cubes has
#							been greatly improved (binCubes is made much faster). 
#           : The group check was moved before the dialog box. 
#						:	The offset options have been removed for the moment. 
#           : Shell mode is now	the default.
#
#   beta 03 : MORE cleanup by Dan Rathbun
#           : Removed "td_" from all method names;
#           : Method shell3d! renamed to shell3d 
#               (because it's not an inplace instance method.)
#           : Wrapped inputbox in a begin .. rescue .. retry block.
#           : Wrapped operation in a begin .. rescue block.
#           : Each operation gets a op number on the Undo stack.
#           : Each cube component definition is now tagged as "VoxBox {#@@op}"
#             and a description is added. This may allow cleaning up unused
#             components.
#           : Added use() method.
#
#		beta 04 : "Spiral algorithm" added to make sure pointy objects are not 
#             interpreted as entrances to the solid.
#           : Methods were rearranged to be more accessible to other plugins.
#           : It's possible to call the voxelize method without having it
#             draw the cubes in Sketchup. This can be used to retrieve 
#             the 3D array, "binCubes" and the array, "cubes", which
#             contains all the cube instances (components).



require "sketchup"

# Check if TheDro namespace is defined
begin
	Module.const_get(:TheDro)
rescue NameError
	module TheDro; end  # define Andrew's TopLevel namespace
end

module TheDro::Voxelize  # define THIS specific plugin submodule
	
	# Thanks to Dan Rathbun for tidying this up!
  unless file_loaded?( File.basename(__FILE__).downcase )

    ###  CONSTANTS
    #
    VERSION = '0.4.0'

    TOPMENU='Plugins'
    SUBMENU=''
    MENUTEXT="Voxelize (#{VERSION})"
    
    Halfscale = Geom::Transformation.scaling 0.5

    ###  MODULE VARS
    #
    @@topmenu = UI.menu(TOPMENU)

    unless SUBMENU.empty?
      @@menu = @@topmenu.add_submenu(SUBMENU)
    else
      @@menu = @@topmenu
    end

    @@item = false
    @@validation = MF_ENABLED
    
    @@cube_size = ""
    
    @@op = 0  ### 0.3.0 added

  end


  class << self

    ### ----------------------------------------------------------------- ###
    #
    #  voxelize()
    #
    def voxelize( draw = true) # I should come up with a better variable name

      model = Sketchup.active_model
      entities = model.entities
      selection = model.selection
      
			group = selection[0]
			if group.class != Sketchup::Group 
        UI.messagebox "No Group Selected!"
        return 
      end

			# OPTIONS INPUTBOX
			title = "Voxelize Options"
			#
      prompts = ["cube size in current units ", "type"]
      defaults = [@@cube_size, "Shell"]
      list = ["", "Shell|Full"]

      begin
        input = UI.inputbox prompts, defaults, list, title

        return [nil,nil] unless input

        
        cubelength = input[0].to_l
				
				raise(ArgumentError) if cubelength == 0

				
      rescue ArgumentError
        ### 0.3.0
        #
        # WAS: UI.messagebox "Invalid Input!"
        title << " (Invalid Cube Size!)" unless title =~ /(Invalid Cube Size!)/
        # in case user changed the "type"
        defaults[1] = input[1]
        # title and defaults will reset the next time voxelize() is called.
        retry # go back to begin and start again.  WAS: return
        #
        ###
      else
        # only if no input errors
        @@cube_size = input[0]
        #
      end # OPTIONS INPUTBOX 

			
			
      ### OPERATION
      begin

        t1 = Time.now

        model.start_operation "Voxelize(cubegrid #{@@op+=1})" ### 0.3.0  added op number



        # Get corners of group
        bound = group.local_bounds
        corners = []
        for ii in 0..7
          corners[ii] = bound.corner(ii)
        end
        
        # Create probing vector (in x direction)
        xVect = Geom::Vector3d.new(corners[0],corners[1])
        nx = (xVect.length/cubelength).round
        xVect.length = cubelength
        
        yVect = Geom::Vector3d.new(corners[0],corners[2])
        ny = (yVect.length/cubelength).round
        yVect.length = cubelength
        
        zVect = Geom::Vector3d.new(corners[0],corners[4])
        nz = (zVect.length/cubelength).round
        zVect.length = cubelength
        
        diagVect = (xVect+yVect+zVect).transform! Halfscale
        diagTrans = (Geom::Transformation.translation(diagVect)).invert!
        
        
        pPointInit = corners[0]
				
        # puts "hello"
        puts "Init: "+(Time.now-t1).to_s
        t1 = Time.now
        # Get all faces
        faces = []
        for ii in group.entities
          if ii.class == Sketchup::Face
            faces << ii
          end
        end
        # puts "Got " + faces.length.to_s + "Faces"
        

        puts "Init Faces: "+(Time.now-t1).to_s
        t1 = Time.now
        
				binCubes = Array.new(nx).map!{ Array.new(ny).map!{ Array.new(nz) } }
        intDist =  Array.new(ny).map!{ Array.new(nz).map!{ Array.new() } }
				
        nyVect = yVect.normalize.to_a
        nzVect = zVect.normalize.to_a
        orig = corners[0].to_a
        frontPlane = [corners[0], xVect]
        for ii in faces
          miny = ny-1
          maxy = 0
          minz = nz-1
          maxz = 0
          
          for vv in ii.vertices
            vertVect = orig.vector_to vv.position
            yDist = nyVect.dot(vertVect)/cubelength
            zDist = nzVect.dot(vertVect)/cubelength
            # puts yDist
            if yDist<miny then miny = yDist end
            if yDist>maxy then maxy = yDist end
            if zDist<minz then minz = zDist end
            if zDist>maxz then maxz = zDist end
            
          end # vv
          
					
					# I don't know how else to cause 0.5 to round down and >0.5 to round up
          
					# miny = ny-(ny-miny).round
          # maxy = maxy.round
          # minz = nz-(nz-minz).round
          # maxz = maxz.round
					
					# Not as efficient as above values but I was runing into 
					# weird missing intersections
					miny = miny.floor
					maxy = maxy.ceil
					minz = minz.floor
					maxz = maxz.ceil
          
          
        
					# puts "miny,maxy,minz,maxz: " + miny.to_s + maxy.to_s + minz.to_s + maxz.to_s
					ns = 50
          for jj in miny..maxy
            yScale = Geom::Transformation.scaling(jj+0.5)
            pPointInit1 = corners[0].offset(yVect.transform(yScale))
            # Sketchup.set_status_text("Filling Solid (" +
              # (jj.to_f/ny*100).round.to_s + "%)") 
          
          
            for kk in minz..maxz
              zScale = Geom::Transformation.scaling(kk+0.5)
              pPointInit2 = pPointInit1.offset(zVect.transform(zScale))
              
              curPoint = Geom.intersect_line_plane([pPointInit2,xVect],
                ii.plane)
              
              
              # Ignore faces parallel to probing vector
							next if curPoint == nil
              
              # # Get array of distances where xVect will cross faces of the solid
              # cPoint = ii.classify_point(curPoint)
              # if ((cPoint == Sketchup::Face::PointInside) or
                  # (cPoint == Sketchup::Face::PointOnVertex) or
                  # (cPoint == Sketchup::Face::PointOnEdge))
              
                # intDist[jj][kk] << pPointInit2.distance(curPoint)
								
								
								
								
							# Check where point is relative to face
							cPoint = ii.classify_point(curPoint)
							# puts curPoint.to_s + cPoint.to_s
							# puts(cPoint.to_s) if cPoint==Sketchup::Face::PointOnEdge
							
							# Point is on face, good
							if (cPoint == Sketchup::Face::PointInside)
							
								intDist[jj][kk] << pPointInit2.distance(curPoint)
								next

							# If point isn't on edge or vertex, go to next
							elsif ((cPoint != Sketchup::Face::PointOnVertex) and
							(cPoint != Sketchup::Face::PointOnEdge))
								
								next
								
							end # if
							
							# If point is on edge or vertex, begin spiral algorithm
							for ss in 1..ns
								
								# shift probing point around a spiral
								sPoint = pPointInit2.offset(yVect.to_a.collect{
									|x| x*ss/ns*Math.cos(Math::PI*2*ss/ns)})
									
								sPoint.offset!(zVect.to_a.collect{
									|x| x*ss/ns*Math.sin(Math::PI*2*ss/ns)})
									
								
								
								curPoint = Geom.intersect_line_plane([sPoint,xVect],
									ii.plane)
								
								# Try to hit the face again
								cPoint = ii.classify_point(curPoint)					
								
								# puts ss.to_s + " : " + curPoint.to_s + cPoint.to_s
								
								# Probing vector hit the face!
								if (cPoint == Sketchup::Face::PointInside)
									
									intDist[jj][kk] << pPointInit2.distance(curPoint)
									break
									
								# Probing vector hit a vertex or edge again, continue spiral
								elsif ((cPoint == Sketchup::Face::PointOnVertex) or
								(cPoint == Sketchup::Face::PointOnEdge))
									
									next
								
								# Probing vector does not intersect face
								else
								
									break
									
								end
								
							end # ss
						end # kk
          end # jj
        end # ii
        
        
        for jj in 0..(ny-1)
          for kk in 0..(nz-1)

            # Solid is not intersected
						next if intDist[jj][kk].empty?
            
						# Sort internal distances
            intDist[jj][kk].sort!

            # If inside the solid, cube should be made
            inside = false
            for ii in 0..(nx-1)
              pos = xVect.length*(ii+0.5)
              
              # Has solid been intersected?
              for nDist in intDist[jj][kk]
                if pos>=nDist
                  intDist[jj][kk] = intDist[jj][kk][1..-1]
                  inside = !inside
                else
                  break
                end
              end # nDist
              # Go to next iteration if not inside solid
              # if not inside then next end
							next if not inside
              # Inside solid: cube should be made
              binCubes[ii][jj][kk] = 1

            end # ii
            
          end # kk
        end # jj

        puts "Get binCubes: "+(Time.now-t1).to_s
        t1 = Time.now
        # Make shell
        ### 0.3.0
        #
        # WAS: if input[1] == "Shell" then shell3d(binCubes) end
        shell3d(binCubes) if input[1] == "Shell"
        #
        ###
        puts "Shell: "+(Time.now-t1).to_s
        t1 = Time.now
				
				cubes = nil
				if draw then
          # Draw cubes
					gTrans = group.transformation
          cubes = drawcubes(binCubes, corners[0],[xVect,yVect,zVect],
            diagTrans, gTrans)
          #puts "hello"
          puts "Draw Cubes: "+(Time.now-t1).to_s
          t1 = Time.now
        end
				
				

        # End of operation
        model.commit_operation

      ### 0.3.0
      #
      rescue Exception => e
        model.abort_operation
        @@op-=1 # because @@op was incremented in the model.start_operation line
        puts("\nVoxelize Operation Error: #<#{e.class.name}: #{e.message}>\n")
        puts(e.backtrace)
        return nil
      else
        # only executes if no errors occur
        puts("\nVoxelize Operation (#{@@op}) Completed at #{Time.now.inspect}.\n")
        return binCubes, cubes
      #ensure
        # ALWAYS executes no matter what happens
      end # operation
      #
      ###

    end # def voxelize()


    ### ----------------------------------------------------------------- ###
    #
    #  cubecomp()
    #
    #  Make a cube component.
    #
    def cubecomp(x,y,z)
      
      model = Sketchup.active_model
      entities = model.entities
      selection = model.selection 
      
      orig = ORIGIN.clone ### WAS: Geom::Point3d.new 0,0,0
      # orig = orig - x.transform(Halfscale) -
        # y.transform(Halfscale) - z.transform(Halfscale)
      pts = []
      pts[0] = orig
      pts[3] = orig + x
      pts[1] = orig + y
      pts[2] = orig + x + y
      
			cubeGroup = entities.add_group
			ents = cubeGroup.entities
      cubeface = ents.add_face pts
      cubeface.pushpull -z.length, true
			cubeface.reverse!
      # cube = cubeface.all_connected
      # group = ents.add_group cube
      cubeinst = cubeGroup.to_component
      cubedef = cubeinst.definition
      ### 0.3
      cubedef.name = "VoxBox #{@@op}"
      cubedef.description = "Cube used by TheDro::Voxelize plugin."
      ###
      entities.erase_entities cubeinst
      return cubedef

    end


    ### ----------------------------------------------------------------- ###
    #
    #  shell3d()
    #
    def shell3d(a)
      
      # puts a.length
      # puts a[0].length
      # puts a[0][0].length
      
      for ii in 1..(a.length-2)
        for jj in 1..(a[0].length-2)
          for kk in 1..(a[0][0].length-2)
            if a[ii][jj][kk] and
            a[ii-1][jj][kk] and a[ii+1][jj][kk] and
            a[ii][jj-1][kk] and a[ii][jj+1][kk] and
            a[ii][jj][kk-1] and a[ii][jj][kk+1] then
              
              a[ii][jj][kk] = 0
            end
          end
        end
      end

    end


    ### ----------------------------------------------------------------- ###
    #
    #  drawcubes()
    #
    def drawcubes(binCubes, pPointInit, vect, diagTrans, gTrans)
      
      model = Sketchup.active_model
      entities = model.entities
      selection = model.selection
      
      cubedef = cubecomp(vect[0],vect[1],vect[2])

      n = [binCubes.length, binCubes[0].length, binCubes[0][0].length]
      cubes = []
      for jj in 0..(n[1]-1)
        # Sketchup.set_status_text("Drawing Cubes (" +
          # (jj.to_f/n[1]*100).round.to_s + "%)") 
        yScale = Geom::Transformation.scaling(jj+0.5)
        pPointInit1 = pPointInit.offset(vect[1].transform(yScale))
      
        for kk in 0..(n[2]-1)
          zScale = Geom::Transformation.scaling(kk+0.5)
          pPointInit2 = pPointInit1.offset(vect[2].transform(zScale))
      
          for ii in 0..(n[0]-1)
            
            if binCubes[ii][jj][kk] == 1
              xScale = Geom::Transformation.scaling(ii+0.5)
              pPoint = pPointInit2.offset(vect[0].transform(xScale))
              trans = Geom::Transformation.translation pPoint.to_a
              cubes << entities.add_instance(cubedef, trans)
              # cubes[-1].transform! diagTrans
              # cubes[-1].transform! gTrans
              
              
            end
          end
        end
      end
      # puts cubes.length.to_s
      entities.transform_entities((gTrans*diagTrans), cubes)
      # entities.transform_entities diagTrans, cubes
      # entities.transform_entities gTrans, cubes

      return cubes  ### 0.3.0  WAS : return

    end
    

    ### ----------------------------------------------------------------- ###
    #
    #  use()
    #
    def use()
      @@validation = MF_CHECKED | MF_GRAYED
      voxelize(true)
      @@validation = MF_ENABLED
    end


  end # class << self


  ### ------------------------------------------------------------------- ###
  #
  #  MENU &| COMMAND
  #
  unless file_loaded?( File.basename(__FILE__).downcase )

    @@item = @@menu.add_item( MENUTEXT ) { use() }

    @@menu.set_validation_proc(@@item) { @@validation }
  
    file_loaded( File.basename(__FILE__).downcase )

  end
	

end # module TheDro::Voxelize


